#!/usr/bin/perl -w
use strict;
use Getopt::Long;

sub usage() {
  print STDERR "Usage: $0 [OPTIONS] <input file(s)>\n";
  print STDERR "Generate test source file for CxxTest.\n";
  print STDERR "\n";
  print STDERR "  -o, --output=NAME    Write output to file NAME\n";
  print STDERR "  -r, --runner=CLASS   Create a main() function that runs CxxTest::CLASS\n";
  print STDERR "  -g, --gui=CLASS      Like --runner, with GUI component\n";
  print STDERR "  --error-printer      Same as --runner=ErrorPrinter\n";
  print STDERR "  --abort-on-fail      Abort tests on failed asserts (like xUnit)\n";
  print STDERR "  --have-eh            Use exception handling (even if not found in tests)\n";
  print STDERR "  --template=TEMPLATE  Use TEMPLATE file to generate the test runner\n";
  exit -1;
}

main();

sub main {
  parseCommandline();
  scanInputFiles();
  writeOutput();
}

#
# Handling the command line
#

my ($output, $runner, $gui, $template, $abortOnFail, $haveStd, $haveEh);

sub parseCommandline() {
  @ARGV = expandWildcards(@ARGV);
  GetOptions( 'output=s'      =>  \$output,
	      'template=s'    => \$template,
	      'runner=s'      => \$runner,
	      'gui=s',        => \$gui,
	      'error-printer' => sub { $runner = "ErrorPrinter"; $haveStd = 1; },
	      'abort-on-fail' => \$abortOnFail,
	      'have-eh'       => \$haveEh,
	      'force'         => sub { print STDERR "cxxtestgen.pl: --force is deprecated\n" }
	    ) || usage();
  scalar @ARGV || usage();

  if ( $gui && !$runner ) {
    $runner = 'StdioPrinter';
  }
}

sub expandWildcards() {
  my @result = ();
  while( my $fn = shift @_ ) {
    push @result, glob($fn);
  }
  return @result;
}

#
# Reading the input files and scanning for test cases
#

my (@suites, $suite, $test, $inBlock);
my $numTotalTests = 0;

sub scanInputFiles() {
  foreach my $file (@ARGV) {
    scanInputFile( $file );
  }
  scalar @suites > 0 or die('No tests defined\n');
}

sub scanInputFile($) {
  my ($file) = @_;
  open FILE, "<$file" or die("Cannot open input file \"$file\"");

  my $line;
  while (defined($line = <FILE>)) {
    scanLineForExceptionHandling( $line );
    scanLineForStandardLibrary( $line );

    scanLineForSuiteStart( $file, $., $line );

    if ( $suite ) {
      if ( lineBelongsToSuite( $suite, $., $line ) ) {
	scanLineForTest( $., $line );
	scanLineForCreate( $., $line );
	scanLineForDestroy( $., $line );
      }
    }
  }
  closeSuite();
  close FILE;
}

sub lineBelongsToSuite($$$) {
  my ($suite, $lineNo, $line) = @_;
  if ( !$suite->{'generated'} ) {
    return 1;
  }

  if ( !$inBlock ) {
    $inBlock = lineStartsBlock( $line );
  }
  if ( $inBlock ) {
    addLineToBlock( $suite->{'file'}, $lineNo, $line );
  }
  return $inBlock;
}

sub scanLineForExceptionHandling($) {
  my ($line) = @_;
  if ( $line =~ m/\b\(throw|try|catch|TS_ASSERT_THROWS|TS_ASSERT_THROWS_ANYTHING|TS_ASSERT_THROWS_NOTHING\)\b/ ) {
    addExceptionHandling();
  }
}

sub scanLineForStandardLibrary($) {
  my ($line) = @_;
  if ( $line =~ m/\b(std\s*::|CXXTEST_STD|using\s+namespace\s+std\b|^\s*\#\s*include\s+<[a-z0-9]+>)/ ) {
    addStandardLibrary();
  }
}

sub scanLineForSuiteStart($$$) {
  my ($fileName, $lineNo, $line) = @_;
  if ( $line =~ m/\bclass\s+(\w+)\s*:\s*public\s+((::)?\s*CxxTest\s*::\s*)?TestSuite\b/ ) {
    startSuite( $1, $fileName, $lineNo, 0 );
  }
  if ( $line =~ m/\bCXXTEST_SUITE\s*\(\s*(\w*)\s*\)/ ) {
    startSuite( $1, $fileName, $lineNo, 1 );
  }
}

sub startSuite($$$$) {
  my ($name, $file, $line, $generated) = @_;
  closeSuite();
  $suite = { 'name' => $name,
	     'file' => $file,
	     'line' => $line,
	     'generated' => $generated,
	     'create' => 0,
	     'destroy' => 0,
	     'tests' => [],
	     'lines' => [] };
}

sub lineStartsBlock($) {
  my ($line) = @_;
  return $line =~ m/\bCXXTEST_CODE\s*\(/;
}

sub scanLineForTest($$) {
  my ($lineNo, $line) = @_;
  if ( $line =~ m/\bvoid\s+([Tt]est\w+)\s*\(\s*(void)?\s*\)/ ) {
    addTest( $1, $lineNo );
  }
}

sub addTest($$$) {
  my ($name, $line) = @_;
  $test = { 'name' => $name,
	    'line' => $line };
  push @{suiteTests()}, $test;
}

sub addLineToBlock($$$) {
  my ($fileName, $lineNo, $line) = @_;
  $line = fixBlockLine( $fileName, $lineNo, $line );
  $line =~ s/^.*\{\{//;
  my $end = ($line =~ s/\}\}.*//s);
  push @{$suite->{'lines'}}, $line;
  if ( $end ) {
    $inBlock = 0;
  }
}

sub fixBlockLine($$$) {
  my ($fileName, $lineNo, $line) = @_;
  my $fileLine = cstr($fileName) . "," . $lineNo;
  $line =~ s/\b(E?TSM?_(ASSERT[A-Z_]*|FAIL))\s*\(/_$1($fileLine,/g;
  return $line;
}

sub scanLineForCreate($$) {
  my ($lineNo, $line) = @_;
  if ( $line =~ m/\bstatic\s+\w+\s*\*\s*createSuite\s*\(\s*(void)?\s*\)/ ) {
    addCreateSuite( $lineNo );
  }
}

sub scanLineForDestroy($$) {
  my ($lineNo, $line) = @_;
  if ( $line =~ m/\bstatic\s+void\s+destroySuite\s*\(\s*\w+\s*\*\s*\w*\s*\)/ ) {
    addDestroySuite( $lineNo );
  }
}

sub closeSuite() {
  if ( $suite && scalar @{suiteTests()} ) {
    verifySuite();
    rememberSuite();
  }
  undef $suite;
}

sub addCreateSuite($) {
  $suite->{'createSuite'} = $_[0];
}

sub addDestroySuite($) {
  $suite->{'destroySuite'} = $_[0];
}

sub addExceptionHandling() {
  $haveEh = 1;
}

sub addStandardLibrary() {
  $haveStd = 1;
}

sub verifySuite() {
  if (suiteCreateLine() || suiteDestroyLine()) {
    die("Suite ", suiteName(), "  must have both createSuite() and destroySuite()\n")
      unless (suiteCreateLine() && suiteDestroyLine());
  }
}

sub rememberSuite() {
  push @suites, $suite;
  $numTotalTests += scalar @{$suite->{'tests'}};
}

sub suiteName() { return $suite->{'name'}; }
sub suiteTests() { return $suite->{'tests'}; }
sub suiteCreateLine() { return $suite->{'createSuite'}; }
sub suiteDestroyLine() { return $suite->{'destroySuite'}; }
sub fileName() { return $suite->{'file'}; }
sub fileString() { return cstr(fileName()); }
sub testName() { return $test->{'name'}; }
sub testLine() { return $test->{'line'}; }

sub suiteObject() { return "suite_".suiteName(); }

sub cstr($) {
  my $file = $_[0];
  $file =~ s/\\/\\\\/g;
  return "\"".$file."\"";
}

#
# Writing the test source file
#

sub writeOutput() {
  $template ? writeTemplateOutput() : writeSimpleOutput();
}

sub startOutputFile() {
  if ( !standardOutput() ) {
    open OUTPUT_FILE,">$output" or die("Cannot create output file \"$output\"");
    select OUTPUT_FILE;
  }
  print "/* Generated file, do not edit */\n\n";
}

sub standardOutput() {
  return !$output;
}

sub writeSimpleOutput() {
  startOutputFile();
  writePreamble();
  writeMain();
  writeWorld();
}

my ($didPreamble, $didWorld);

sub writeTemplateOutput() {
  openTemplateFile();
  startOutputFile();
  my $line;
  while (defined($line = <TEMPLATE_FILE>)) {
    if ( $line =~ m/^\s*\#\s*include <cxxtest\// ) {
      writePreamble();
      print $line;
    } elsif ( $line =~ m/^\s*<CxxTest\s+world>\s*$/ ) {
      writeWorld();
    } else {
      print $line;
    }
  }
}

sub openTemplateFile() {
  open TEMPLATE_FILE, "<$template" or die("Cannot open template file \"$template\"");
}

sub writePreamble() {
  return if $didPreamble;
  if ( $haveStd ) {
    print "#ifndef CXXTEST_HAVE_STD\n";
    print "#define CXXTEST_HAVE_STD\n";
    print "#endif\n";
  }
  if ( $haveEh ) {
    print "#ifndef CXXTEST_HAVE_EH\n";
    print "#define CXXTEST_HAVE_EH\n";
    print "#endif\n";
  }
  if ( $abortOnFail ) {
    print "#ifndef CXXTEST_ABORT_TEST_ON_FAIL\n";
    print "#define CXXTEST_ABORT_TEST_ON_FAIL\n";
    print "#endif\n";
  }
  print "#ifndef CXXTEST_RUNNING\n";
  print "#define CXXTEST_RUNNING\n";
  print "#endif\n";
  print "\n";
  print "#include <cxxtest/TestListener.h>\n";
  print "#include <cxxtest/CountingTracker.h>\n";
  print "#include <cxxtest/TestRunner.h>\n";
  print "#include <cxxtest/$runner.h>\n" if $runner;
  print "#include <cxxtest/$gui.h>\n" if $gui;
  print "\n";
  print "typedef const CxxTest::SuiteDescription *SuiteDescriptionPtr;\n";
  print "typedef const CxxTest::TestDescription *TestDescriptionPtr;\n";
  print "\n";
  $didPreamble = 1;
}

sub writeWorld() {
  return if $didWorld;
  writePreamble();
  writeSuites();
  writeWorldDescription();
  writeClassStatics();
  $didWorld = 1;
}

sub writeSuites() {
  foreach (@suites) {
    $suite = $_;
    includeFile(fileName());
    if ( $suite->{'generated'} ) { generateSuite(); }
    dynamicSuite() ? writeSuitePointer() : writeSuiteObject();
    writeTestDescriptionsBase();
    writeTestDescriptions();
    writeTestPointers();
    writeSuiteDescription();
  }
  writeSuitePointers();
}

sub dynamicSuite() {
  return suiteCreateLine();
}

my $lastIncluded;

sub includeFile($) {
  my $file = $_[0];
  return if $lastIncluded && ($file eq $lastIncluded);
  print "#include \"$file\"\n\n";
  $lastIncluded = $file;
}

sub generateSuite() {
  print "class ", suiteName(), " : public CxxTest::TestSuite {\n";
  print "public:\n";
  foreach my $line (@{$suite->{'lines'}}) {
    print $line;
  }
  print "};\n\n";
}

sub writeTestDescriptionsBase() {
  print "class TestDescriptionBase_", suiteName(), " : public CxxTest::TestDescription {\n";
  print "public:\n";
  print " const char *file() const { return ", fileString(), "; }\n";
  print " const char *suiteName() const { return \"", suiteName(), "\"; }\n";
  print "};\n\n";
}

sub writeTestDescriptions() {
  foreach (@{suiteTests()}) {
    $test = $_;
    writeTestDescription();
  }
}

sub writeTestPointers() {
  print "static const TestDescriptionPtr testDescriptions_", suiteName(), "[] = {\n";
  foreach (@{suiteTests()}) {
    $test = $_;
    print " &testDescription_", suiteName(), "_", testName(), ",\n";
  }
  print " &CxxTest::TestDescription::_dummy\n";
  print "};\n\n";
}

sub writeSuitePointer() {
  print "static ", suiteName(), " *", suiteObject(), " = 0;\n\n";
}

sub writeSuiteObject() {
  print "static ", suiteName(), " ", suiteObject(), ";\n\n";
}

sub writeTestDescription() {
  print "static class TestDescription_", suiteName(), "_", testName(), " : public TestDescriptionBase_", suiteName(), " {\n";
  print "public:\n";
  print " unsigned line() const { return ", testLine(), "; }\n";
  print " const char *testName() const { return \"", testName(), "\"; }\n";
  dynamicSuite() ? writeDynamicRun() : writeStaticRun();
  print "} testDescription_", suiteName(), "_", testName(), ";\n\n";
}

sub writeDynamicRun() {
  print " void run() const { if ( ", suiteObject(), " ) ", suiteObject(), "->", testName(), "(); }\n";
}

sub writeStaticRun() {
  print " void run() const { ", suiteObject(), ".", testName(), "(); }\n";
}

sub writeSuiteDescription() {
  print "static class SuiteDescription_", suiteName(), " : public CxxTest::SuiteDescription {\n";
  print "public:\n";
  print " const char *file() const { return ", fileString(), "; }\n";
  print " unsigned line() const { return ", $suite->{'line'}, "; }\n";
  print " const char *suiteName() const { return \"", suiteName(), "\"; }\n";
  print " unsigned numTests() const { return ", scalar @{suiteTests()}, "; }\n";
  dynamicSuite() ? writeReturnDynamicSuite() : writeReturnStaticSuite();
  print " const CxxTest::TestDescription &testDescription( unsigned i ) const ",
    "{ return *(testDescriptions_", suiteName(), "[i]); }\n";
  print "} suiteDescription_", suiteName(), ";\n\n";
}

sub writeReturnDynamicSuite() {
  print " CxxTest::TestSuite *suite() const { return ", suiteObject(), "; }\n";
}

sub writeReturnStaticSuite() {
  print " CxxTest::TestSuite *suite() const { return &", suiteObject(), "; }\n";
}

sub writeSuitePointers() {
  print "static const SuiteDescriptionPtr suiteDescriptions_TheWorld[] = {\n";
  foreach (@suites) {
    print " &suiteDescription_", $_->{'name'}, ",\n";
  }
  print " &CxxTest::SuiteDescription::_dummy\n";
  print "};\n\n";
}

sub writeWorldDescription() {
  print "static class WorldDescription_TheWorld : public CxxTest::WorldDescription {\n";
  print "public:\n";
  print " unsigned numSuites() const { return ", scalar @suites, "; }\n";
  print " unsigned numTotalTests() const { return ", $numTotalTests, "; }\n";
  print " const CxxTest::SuiteDescription &suiteDescription( unsigned i ) const\n";
  print "  { return *(suiteDescriptions_TheWorld[i]); }\n";
  print "protected:\n";
  writeWorldSetUp();
  writeWorldTearDown();
  print "} worldDescription_theWorld;\n\n";
}

sub writeWorldSetUp() {
  print " void setUp() const {\n";
  foreach (@suites) {
    writeCreateSuite($_);
  }
  print " }\n";
}

sub writeCreateSuite($) {
  $suite = $_[0];
  return unless dynamicSuite();
  print "  __TS_ASSERT_THROWS_NOTHING( ", fileString(), ", ", suiteCreateLine(), ",\n",
        "                              ", suiteObject(), " = ", suiteName(), "::createSuite(),\n",
	"                              \"", suiteName(), "::createSuite()\" );\n";
  print "  __TS_ASSERT( ",  fileString(), ", ", suiteCreateLine(), ", ", suiteObject(), " != 0,\n",
        "               \"", suiteName(), "::createSuite() != 0\" );\n";
}

sub writeWorldTearDown() {
  print " void tearDown() const {\n";
  foreach (@suites) {
    writeDestroySuite($_);
  }
  print " }\n";
}

sub writeDestroySuite($) {
  $suite = $_[0];
  return unless dynamicSuite();
  print "  if ( ", suiteObject(), " )\n";
  print "   __TS_ASSERT_THROWS_NOTHING( ", fileString(), ", ", suiteDestroyLine(), ",\n",
        "                               ", suiteName(), "::destroySuite( ", suiteObject(), " ),\n",
	"                               \"", suiteName(), "::destroySuite()\" );\n";
}

sub writeClassStatics() {
  print "CxxTest::TestListener CxxTest::TestListener::_dummy;\n";
  print "CxxTest::TestTracker CxxTest::TestTracker::_dummy;\n";
  print "const CxxTest::WorldDescription CxxTest::WorldDescription::_dummy;\n";
  print "const CxxTest::SuiteDescription CxxTest::SuiteDescription::_dummy;\n";
  print "const CxxTest::TestDescription CxxTest::TestDescription::_dummy;\n";
  print "CxxTest::TestTracker *CxxTest::TestTracker::_tracker = &CxxTest::TestTracker::_dummy;\n";
  print "CxxTest::TestListener *CxxTest::TestListener::_listener = &CxxTest::TestListener::_dummy;\n";
  print "\n";
  print "CxxTest::CountingTracker CxxTest::TestRunner::_counter;\n";
  print "const CxxTest::WorldDescription &CxxTest::TestRunner::_world = worldDescription_theWorld;\n\n";
}

sub writeMain() {
  if ( $gui ) {
    print "int main( int argc, char *argv[] ) {\n";
    print " return CxxTest::GuiTuiRunner<CxxTest::$gui, CxxTest::$runner>( argc, argv ).run();\n";
    print "}\n";
  }
  elsif ( $runner ) {
    print "int main( int argc, char *argv[] ) {\n";
    print " return CxxTest::$runner().run();\n";
    print "}\n";
  }
}
