#ifndef __CXXTEST__ERRORFORMATTER_H
#define __CXXTEST__ERRORFORMATTER_H

//
// The ErrorFormatter is a TestListener that
// prints reports of the errors to an output
// stream.  Since we cannot rely ou the standard
// iostreams, this header defines a base class
// analogout to std::ostream.
//

#include <cxxtest/TestRunner.h>
#include <cxxtest/TestListener.h>
#include <cxxtest/TestTracker.h>

namespace CxxTest
{
    class OutputStream
    {
    public:
        virtual ~OutputStream() {}
        virtual void flush() {};
        virtual OutputStream &operator<<( unsigned i ) { return *this; }
        virtual OutputStream &operator<<( const char *s ) { return *this; }
        OutputStream &operator<<( void (*f)( OutputStream & ) ) { f( *this ); return *this; }
        static void endl( OutputStream &o ) { (o << "\n").flush(); }
    };
    
    class ErrorFormatter : public TestListener
    {
    public:
        ErrorFormatter( OutputStream *o, const char *preLine = ":", const char *postLine = "" ) :
            _o(o), _preLine(preLine), _postLine(postLine) {}

        int run()
        {
            TestRunner::runAllTests(this);
            return tracker().failedTests();
        }

	const TestTracker &tracker() const
        {
            return TestTracker::tracker();
        }

	void enterWorld( const WorldDescription &desc )
	{
	    (*_o) << "Running " << desc.numTotalTests() << " test(s)";
            _o->flush();
	    _dotting = true;
	}

        void enterTest( const TestDescription &desc )
        {
            _reported = false;
        }

	void leaveTest( const TestDescription &desc )
	{
	    if ( !tracker().testFailed() ) {
		((*_o) << ".").flush();
		_dotting = true;
	    }
	}

	void leaveWorld( const WorldDescription &desc )
	{
	    if ( !tracker().failedTests() ) {
		(*_o) << "OK!" << OutputStream::endl;
		return;
	    }
	    newLine();
	    (*_o) << "Failed " << tracker().failedTests() << " of " << desc.numTotalTests() << " test(s)" << OutputStream::endl;
	    unsigned numPassed = desc.numTotalTests() - tracker().failedTests();
	    (*_o) << "Success rate: " << (numPassed * 100 / desc.numTotalTests()) << "%" << OutputStream::endl;
	}

        void warning( const char *file, unsigned line, const char *expression )
        {
            stop( file, line ) << "Warning: " <<
                expression << OutputStream::endl;
        }

        void failedTest( const char *file, unsigned line, const char *expression )
        {
            stop( file, line ) << "Test failed: " <<
                expression << OutputStream::endl;
        }
    
	void failedAssert( const char *file, unsigned line, const char *expression )
	{
	    stop( file, line ) << "Assertion failed: " <<
		expression << OutputStream::endl;
	}

	void failedAssertEquals( const char *file, unsigned line,
				 const char *xStr, const char *yStr,
				 const char *x, const char *y )
	{
	    stop( file, line ) << "Expected (" <<
		xStr << " == " << yStr << "), found (" <<
		x << " != " << y << ")" << OutputStream::endl;
	}

	void failedAssertDelta( const char *file, unsigned line,
                                const char *xStr, const char *yStr, const char *dStr,
                                const char *x, const char *y, const char *d )
	{
	    stop( file, line ) << "Expected (" <<
		xStr << " == " << yStr << ") up to " << dStr << " (" << d << "), found (" <<
		x << " != " << y << ")" << OutputStream::endl;
	}

	void failedAssertDiffers( const char *file, unsigned line,
                                  const char *xStr, const char *yStr,
                                  const char *value )
	{
	    stop( file, line ) << "Expected (" <<
		xStr << " != " << yStr << "), found (" <<
		value << ")" << OutputStream::endl;
	}

	void failedAssertLessThan( const char *file, unsigned line,
				   const char *xStr, const char *yStr,
				   const char *x, const char *y )
	{
	    stop( file, line ) << "Expected (" <<
		xStr << " < " << yStr << "), found (" <<
		x << " >= " << y << ")" << OutputStream::endl;
	}

	void failedAssertThrows( const char *file, unsigned line,
				 const char *expression, const char *type,
				 bool otherThrown )
        {
            stop( file, line ) << "Expected (" << expression << ") to throw (" <<
                type << ") but it " << (otherThrown ? "threw something else" : "didn't throw") <<
                OutputStream::endl;
        }
        
	void failedAssertThrowsNot( const char *file, unsigned line, const char *expression )
        {
            stop( file, line ) << "Expected (" << expression << ") not to throw, but it did" <<
                OutputStream::endl;
        }

    protected:        
        OutputStream *outputStream() const
        {
            return _o;
        }
        
    private:
	OutputStream &stop( const char *file, unsigned line )
	{
	    newLine();
	    reportTest();
            return (*_o) << file << _preLine << line << _postLine << ": ";
	}
    
	void newLine( void )
	{
	    if ( _dotting ) {
		(*_o) << OutputStream::endl;
		_dotting = false;
	    }
	}

	void reportTest( void )
	{
            if( _reported )
                return;
            (*_o) << "In " << tracker().suite().suiteName() << "::" << tracker().test().testName() << ":" << OutputStream::endl;
            _reported = true;
	}

	bool _dotting;
        bool _reported;
        OutputStream *_o;
        const char *_preLine;
        const char *_postLine;
    };
};

#endif // __CXXTEST__ERRORFORMATTER_H
