#ifndef __CXXTEST__QTGUI_H
#define __CXXTEST__QTGUI_H

//
// The QtGui displays a simple progress bar using the Qt Toolkit
// Apart from normal Qt command-line arguments, it accepts -minimized
//

#include <cxxtest/Gui.h>

#include <qapplication.h>
#include <qglobal.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qmessagebox.h>
#include <qpixmap.h>
#include <qprogressbar.h>
#include <qstatusbar.h>

namespace CxxTest
{
    class QtGui : public GuiListener
    {
    public:
	void enterGui( int &argc, char **argv )
	{
            parseCommandLine( argc, argv );
            createApplication( argc, argv );
        }

        void enterWorld( unsigned numTotalTests )
        {
	    createWindow( numTotalTests );
	    processEvents();
	}

	void enterSuite( const SuiteDescription &d )
	{
	    showSuiteName( d.suiteName() );
	}

        void enterTest( const char *suiteName, const char *testName )
        {
            setCaption( suiteName, testName );
            advanceProgressBar();
	    showTestName( testName );
	    showTestsDone( _progressBar->progress() );
	    processEvents();
        }

        void yellowBar()
        {
	    setColor( 255, 255, 0 );
	    setIcon( QMessageBox::Warning );
	    processEvents();
        }
        
        void redBar()
        {
            if ( _startMinimized && _mainWindow->isMinimized() )
                showNormal();
	    setColor( 255, 0, 0 );
	    setIcon( QMessageBox::Critical );
	    processEvents();
        }

	void leaveGui()
	{
            _mainWindow->close( true );
        }

    private:
	QString _programName;
        bool _startMinimized;
	unsigned _numTotalTests;
	QApplication *_application;
        QWidget *_mainWindow;
        QVBoxLayout *_layout;
	QProgressBar *_progressBar;
        QStatusBar *_statusBar;
        QLabel *_suiteName, *_testName, *_testsDone;

        void parseCommandLine( int argc, char **argv )
        {
            _startMinimized = false;
	    _programName = argv[0];
            for ( int i = 1; i < argc; ++ i )
                if ( QString( argv[i] ) == "-minimized" )
                    _startMinimized = true;
        }

        void createApplication( int &argc, char **argv )
        {
	    _application = new QApplication( argc, argv );
	}	
        
	void createWindow( unsigned numTotalTests )
	{
	    _numTotalTests = numTotalTests;
            createMainWindow();
	    createProgressBar();
            createStatusBar();
            setMainWidget();
            if ( _startMinimized )
                showMinimized();
            else
                showNormal();
	}

        void createMainWindow()
        {
            _mainWindow = new QWidget();
            _layout = new QVBoxLayout( _mainWindow );
        }

	void createProgressBar()
	{
            _layout->addWidget( _progressBar = new QProgressBar( _numTotalTests, _mainWindow ) );
	    _progressBar->setProgress( 0 );
	    setColor( 0, 255, 0 );
	    setIcon( QMessageBox::Information );
	}

        void createStatusBar()
        {
            _layout->addWidget( _statusBar = new QStatusBar( _mainWindow ) );
            _statusBar->addWidget( _suiteName = new QLabel( _statusBar ), 2 );
            _statusBar->addWidget( _testName = new QLabel( _statusBar ), 4 );
            _statusBar->addWidget( _testsDone = new QLabel( _statusBar ), 1 );
        }

        void setMainWidget()
        {
	    _application->setMainWidget( _mainWindow );
        }

        void showMinimized()
        {
            _mainWindow->showMinimized();
        }

	void showNormal()
	{
            _mainWindow->showNormal();
            centerWindow();
        }

        void setCaption( const QString &suiteName, const QString &testName )
        {
	    _mainWindow->setCaption( _programName + " - " + suiteName + "::" + testName + "()" );
        }

	void showSuiteName( const QString &suiteName )
	{
            _suiteName->setText( "class " + suiteName );
	}

        void advanceProgressBar()
        {
            _progressBar->setProgress( _progressBar->progress() + 1 );
        }

        void showTestName( const QString &testName )
        {
            _testName->setText( testName + "()" );
	}

	void showTestsDone( unsigned testsDone )
	{
            _testsDone->setText( asString( testsDone ) + " of " + asString( _numTotalTests ) );
        }

	static QString asString( unsigned n )
	{
	    return QString::number( n );
	}

	void setColor( int r, int g, int b )
	{
	    QPalette palette = _progressBar->palette();
            palette.setColor( QColorGroup::Highlight, QColor( r, g, b ) );
	    _progressBar->setPalette( palette );
	}

	void setIcon( QMessageBox::Icon icon )
	{
#if QT_VERSION >= 0x030000
	    _mainWindow->setIcon( QMessageBox::standardIcon( icon ) );
#else // Qt version < 3.0.0
	    _mainWindow->setIcon( QMessageBox::standardIcon( icon, QApplication::style().guiStyle() ) );
#endif // QT_VERSION
	}

	void processEvents()
	{
	    _application->processEvents();
	}

	void centerWindow()
	{
	    QWidget *desktop = QApplication::desktop();
	    int xCenter = desktop->x() + (desktop->width() / 2);
	    int yCenter = desktop->y() + (desktop->height() / 2);
	    
	    int windowWidth = (desktop->width() * 4) / 5;
	    int windowHeight = _mainWindow->height();
	    _mainWindow->setGeometry( xCenter - (windowWidth / 2), yCenter - (windowHeight / 2), windowWidth, windowHeight );
	}
    };
};

#endif // __CXXTEST__QTGUI_H
