#ifndef __CXXTEST_TESTRUNNER_H
#define __CXXTEST_TESTRUNNER_H

//
// TestRunner is the class that runs all the tests.
// To use it, create an object that implements the TestListener
// interface and call TestRunner::runAllTests( &myListener );
// 

#include <cxxtest/TestListener.h>
#include <cxxtest/Descriptions.h>
#include <cxxtest/TestSuite.h>
#include <cxxtest/CountingTracker.h>

namespace CxxTest 
{
#   if defined(CXXTEST_HAVE_EH) && defined(CXXTEST_ABORT_TEST_ON_FAIL)
#       define RUNNER_TRY           _TS_TRY
#       define RUNNER_CATCH(f,l,m)  catch(CxxTest::TestFailed) {} ___TSM_CATCH(f,l,m)
#   else
#       define RUNNER_TRY           _TS_TRY
#       define RUNNER_CATCH         __TSM_CATCH
#   endif
    
    class TestRunner
    {
    public:
        static void runAllTests( TestListener *l )
        {
            RUNNER_TRY { TestRunner(l).runWorld( _world ); } RUNNER_CATCH(__FILE__, __LINE__, "Exception thrown from world");
        }
    
    private:
        static const WorldDescription &_world;
        static CountingTracker _counter;

        TestRunner( TestListener *finalListener )
        {
            _counter.setListener( finalListener );
            TestListener::setListener( &_counter );
            TestTracker::setTracker( &_counter );
        }

        void runWorld( const WorldDescription &wd )
        {
            _counter.enterWorld( wd );
            wd.setUp();
            for ( unsigned i = 0; i < wd.numSuites(); ++ i )
                if ( wd.suiteDescription(i).suite() )
                    runSuite( wd.suiteDescription(i) );
            wd.tearDown();
            _counter.leaveWorld( wd );
        }
    
        void runSuite( const SuiteDescription &sd )
        {
            _counter.enterSuite( sd );
            for ( unsigned i = 0; i < sd.numTests(); ++ i ) {
                RUNNER_TRY { sd.suite()->setUp(); } RUNNER_CATCH(sd.file(), sd.line(), "Exception thrown from setUp()");
                runTest( sd.testDescription(i) );
                RUNNER_TRY { sd.suite()->tearDown(); } RUNNER_CATCH(sd.file(), sd.line(), "Exception thrown from tearDown()");
            }
            _counter.leaveSuite( sd );
        }

        void runTest( const TestDescription &td )
        {
            _counter.enterTest( td );
            RUNNER_TRY { td.run(); } RUNNER_CATCH(td.file(), td.line(), "Exception thrown from test");
            _counter.leaveTest( td );
        }
    };

#   undef RUNNER_TRY
#   undef RUNNER_CATCH
};


#endif // __CXXTEST_TESTRUNNER_H
