#ifndef __CXXTEST__TESTSUITE_H
#define __CXXTEST__TESTSUITE_H

//
// class TestSuite is the base class for all test suites.
// To define a test suite, derive from this class and add
// member functions called void test*();
//

#include <cxxtest/TestListener.h>
#include <cxxtest/Descriptions.h>
#include <cxxtest/ValueTraits.h>
#ifdef CXXTEST_HAVE_STD
#include <stdexcept>
#endif // CXXTEST_HAVE_STD

namespace CxxTest 
{
    class TestSuite
    {
    public:
        virtual ~TestSuite() {}
	virtual void setUp() {}
	virtual void tearDown() {}
    };

#   ifdef CXXTEST_HAVE_EH
    class TestFailed {};
#       define _TS_TRY try
#       ifdef CXXTEST_HAVE_STD
#           define ___TSM_CATCH(f,l,m) \
                    catch(const std::exception &e) { (CxxTest::TestListener::listener()).failedTest(f,l,e.what()); } \
                    catch(...) { (CxxTest::TestListener::listener()).failedTest(f,l,m); }
#       else // !CXXTEST_HAVE_STD
#           define ___TSM_CATCH(f,l,m) \
                    catch(...) { (CxxTest::TestListener::listener()).failedTest(f,l,m); }
#       endif // CXXTEST_HAVE_STD
#       define __TSM_CATCH(f,l,m) \
                catch(CxxTest::TestFailed) { throw; } \
                ___TSM_CATCH(f,l,m)
#       define __TS_CATCH(f,l) __TSM_CATCH(f,l,"Unhandled exception")
#       define _TS_CATCH __TS_CATCH(__FILE__,__LINE__)
#   else // !CXXTEST_HAVE_EH
#       define _TS_TRY
#       define __TSM_CATCH(f,l,m)
#       define __TS_CATCH(f,l)
#       define _TS_CATCH
#   endif // CXXTEST_HAVE_EH

#   define TS_WARN(e) { (CxxTest::TestListener::listener()).warning(__FILE__, __LINE__, TS_AS_STRING(e)); }

#   if defined(CXXTEST_HAVE_EH) && defined(CXXTEST_ABORT_TEST_ON_FAIL)
#       define TS_ABORT() throw CxxTest::TestFailed()
#   else
#       define TS_ABORT()
#   endif
    
#   define _TS_FAIL(f,l,e) { (CxxTest::TestListener::listener()).failedTest(f, l, TS_AS_STRING(e)); TS_ABORT(); }
#   define TS_FAIL(e) _TS_FAIL(__FILE__, __LINE__, e)

#   define __ETS_ASSERT(f,l,expr,msg) { if ( !(expr) ) { \
	    (CxxTest::TestListener::listener()).failedAssert(f, l,msg); TS_ABORT(); } }
#   define __TS_ASSERT(f,l,expr,msg) { _TS_TRY { __ETS_ASSERT(f,l,expr,msg); } __TS_CATCH(f,l) }

#   define _ETS_ASSERT(f,l,e) __ETS_ASSERT(f,l,e,#e)
#   define _TS_ASSERT(f,l,e) __TS_ASSERT(f,l,e,#e)
	
#   define ETS_ASSERT(e) _ETS_ASSERT(__FILE__,__LINE__,e)
#   define TS_ASSERT(e) _TS_ASSERT(__FILE__,__LINE__,e)
        
#   define _ETS_ASSERT_EQUALS(f,l,x,y) { if ( !((x) == (y)) ) { \
            CxxTest::TestListener::listener().failedAssertEquals(f,l,#x,#y,TS_AS_STRING(x),TS_AS_STRING(y)); TS_ABORT(); } }
#   define _TS_ASSERT_EQUALS(f,l,x,y) { _TS_TRY { _ETS_ASSERT_EQUALS(f,l,x,y); } __TS_CATCH(f,l) }
        
#   define ETS_ASSERT_EQUALS(x,y) _ETS_ASSERT_EQUALS(__FILE__,__LINE__,x,y)
#   define TS_ASSERT_EQUALS(x,y) _TS_ASSERT_EQUALS(__FILE__,__LINE__,x,y)
        
#   define _ETS_ASSERT_DIFFERS(f,l,x,y) { if ( ((x) == (y)) ) { \
            CxxTest::TestListener::listener().failedAssertDiffers(f,l,#x,#y,TS_AS_STRING(x)); TS_ABORT(); } }
#   define _TS_ASSERT_DIFFERS(f,l,x,y) { _TS_TRY { _ETS_ASSERT_DIFFERS(f,l,x,y); } __TS_CATCH(f,l) }
        
#   define ETS_ASSERT_DIFFERS(x,y) _ETS_ASSERT_DIFFERS(__FILE__,__LINE__,x,y)
#   define TS_ASSERT_DIFFERS(x,y) _TS_ASSERT_DIFFERS(__FILE__,__LINE__,x,y)
        
#   define _ETS_ASSERT_LESS_THAN(f,l,x,y) { if ( !((x) < (y)) ) { \
            CxxTest::TestListener::listener().failedAssertLessThan(f,l,#x,#y,TS_AS_STRING(x),TS_AS_STRING(y)); TS_ABORT(); } }
#   define _TS_ASSERT_LESS_THAN(f,l,x,y) { _TS_TRY { _ETS_ASSERT_LESS_THAN(f,l,x,y); } __TS_CATCH(f,l) }
        
#   define ETS_ASSERT_LESS_THAN(x,y) _ETS_ASSERT_LESS_THAN(__FILE__,__LINE__,x,y)
#   define TS_ASSERT_LESS_THAN(x,y) _TS_ASSERT_LESS_THAN(__FILE__,__LINE__,x,y)
        
#   define _ETS_ASSERT_DELTA(f,l,x,y,d)  { if ( ((y) < (x) - (d)) || ((y) > (x) + (d)) ) { \
            CxxTest::TestListener::listener().failedAssertDelta(f,l,#x,#y,#d,TS_AS_STRING(x),TS_AS_STRING(y),TS_AS_STRING(d)); \
            TS_ABORT(); } }
#   define _TS_ASSERT_DELTA(f,l,x,y,d) { _TS_TRY { _ETS_ASSERT_DELTA(f,l,x,y,d); } __TS_CATCH(f,l) }
        
#   define ETS_ASSERT_DELTA(x,y,d) _ETS_ASSERT_DELTA(__FILE__,__LINE__,x,y,d)
#   define TS_ASSERT_DELTA(x,y,d) _TS_ASSERT_DELTA(__FILE__,__LINE__,x,y,d)
        
#   define _TS_ASSERT_THROWS(f,l,e,t) { \
            bool _ts_threw_expected = false, _ts_threw_else = false; \
            try { e; } \
            catch( t ) { _ts_threw_expected = true; } \
            catch( ... ) { _ts_threw_else = true; } \
            if ( !_ts_threw_expected ) { \
                (CxxTest::TestListener::listener()).failedAssertThrows(f, l, #e, #t, _ts_threw_else ); \
                TS_ABORT(); \
            } }

#   define TS_ASSERT_THROWS(e,t) _TS_ASSERT_THROWS(__FILE__,__LINE__,e,t)
        
#   define _TS_ASSERT_THROWS_ANYTHING(f,l,e) { \
            bool _ts_threw = false; \
            try { e; } \
            catch( ... ) { _ts_threw = true; } \
            if ( !_ts_threw ) { \
                (CxxTest::TestListener::listener()).failedAssertThrows(f, l, #e, "...", false ); \
                TS_ABORT(); \
            } }

#   define TS_ASSERT_THROWS_ANYTHING(e) _TS_ASSERT_THROWS_ANYTHING(__FILE__, __LINE__, e)
        
#   define __TS_ASSERT_THROWS_NOTHING(f,l,expr,msg) { \
            try { expr; } \
            catch( ... ) { \
                (CxxTest::TestListener::listener()).failedAssertThrowsNot(f, l, msg); } }

#   define _TS_ASSERT_THROWS_NOTHING(f,l,e) __TS_ASSERT_THROWS_NOTHING(f,l,e,#e)

#   define TS_ASSERT_THROWS_NOTHING(e) _TS_ASSERT_THROWS_NOTHING(__FILE__,__LINE__,e)

#   define _ETSM_ASSERT(f,l,m,e) { if ( !(e) ) _TS_FAIL(f,l,m); }
#   define _TSM_ASSERT(f,l,m,e) { _TS_TRY { _ETSM_ASSERT(f,l,m,e); } __TS_CATCH(f,l) }
        
#   define ETSM_ASSERT(m,e) _ETSM_ASSERT(__FILE__,__LINE__,m,e)
#   define TSM_ASSERT(m,e) _TSM_ASSERT(__FILE__,__LINE__,m,e)
        
#   define _ETSM_ASSERT_EQUALS(f,l,m,x,y) { if ( !((x) == (y)) ) _TS_FAIL(f,l,m); }
#   define _TSM_ASSERT_EQUALS(f,l,m,x,y) { _TS_TRY { _ETSM_ASSERT_EQUALS(f,l,m,x,y); } __TS_CATCH(f,l) }
        
#   define ETSM_ASSERT_EQUALS(m,x,y) _ETSM_ASSERT_EQUALS(__FILE__,__LINE__,m,x,y)
#   define TSM_ASSERT_EQUALS(m,x,y) _TSM_ASSERT_EQUALS(__FILE__,__LINE__,m,x,y)
        
#   define _ETSM_ASSERT_DELTA(f,l,m,x,y,d) { if ( ((y) < (x) - (d)) || ((y) > (x) + (d)) ) _TS_FAIL(f,l,m); }
#   define _TSM_ASSERT_DELTA(f,l,m,x,y,d) { _TS_TRY { _ETSM_ASSERT_DELTA(f,l,m,x,y,d); } __TS_CATCH(f,l) }
        
#   define ETSM_ASSERT_DELTA(m,x,y,d) _ETSM_ASSERT_DELTA(__FILE__,__LINE__,m,x,y,d)
#   define TSM_ASSERT_DELTA(m,x,y,d) _TSM_ASSERT_DELTA(__FILE__,__LINE__,m,x,y,d)
        
#   define _ETSM_ASSERT_DIFFERS(f,l,m,x,y) { if ( ((x) == (y)) ) _TS_FAIL(f,l,m); }
#   define _TSM_ASSERT_DIFFERS(f,l,m,x,y) { _TS_TRY { _ETSM_ASSERT_DIFFERS(f,l,m,x,y); } __TS_CATCH(f,l) }
        
#   define ETSM_ASSERT_DIFFERS(m,x,y) _ETSM_ASSERT_DIFFERS(__FILE__,__LINE__,m,x,y)
#   define TSM_ASSERT_DIFFERS(m,x,y) _TSM_ASSERT_DIFFERS(__FILE__,__LINE__,m,x,y)
        
#   define _ETSM_ASSERT_LESS_THAN(f,l,m,x,y) { if ( !((x) < (y)) ) _TS_FAIL(f,l,m); }
#   define _TSM_ASSERT_LESS_THAN(f,l,m,x,y) { _TS_TRY { _ETSM_ASSERT_LESS_THAN(f,l,m,x,y); } __TS_CATCH(f,l) }
        
#   define ETSM_ASSERT_LESS_THAN(m,x,y) _ETSM_ASSERT_LESS_THAN(__FILE__,__LINE__,m,x,y)
#   define TSM_ASSERT_LESS_THAN(m,x,y) _TSM_ASSERT_LESS_THAN(__FILE__,__LINE__,m,x,y)
        
#   define _TSM_ASSERT_THROWS(f,l,m,e,t) { \
            bool _ts_threw = false; \
            try { e; } \
            catch( t ) { _ts_threw = true; } \
            catch( ... ) {} \
            if ( !_ts_threw ) _TS_FAIL(f,l,m); }

#   define TSM_ASSERT_THROWS(m,e,t) _TSM_ASSERT_THROWS(__FILE__,__LINE__,m,e,t)
        
#   define _TSM_ASSERT_THROWS_ANYTHING(f,l,m,e) { \
            bool _ts_threw = false; \
            try { e; } \
            catch( ... ) { _ts_threw = true; } \
            if ( !_ts_threw ) _TS_FAIL(f,l,m); }

#   define TSM_ASSERT_THROWS_ANYTHING(m,e) _TSM_ASSERT_THROWS_ANYTHING(__FILE__,__LINE__,m,e)
        
#   define _TSM_ASSERT_THROWS_NOTHING(f,l,m,e) { try { e; } catch( ... ) { _TS_FAIL(f,l,m); } }

#   define TSM_ASSERT_THROWS_NOTHING(m,e) _TSM_ASSERT_THROWS_NOTHING(__FILE__,__LINE__,m,e)
}



#endif // __CXXTEST__TESTSUITE_H
