#ifndef __CXXTEST__VALUETRAITS_H
#define __CXXTEST__VALUETRAITS_H

//
// ValueTraits are used by CxxTest to convert arbitrary
// values used in TS_ASSERT_EQUALS() to a string representation.
// 
// This header file contains value traits for builtin integral types.
// To declare value traits for new types you should instantiate the class
// ValueTraits<YourClass>.
//

#ifdef CXXTEST_OLD_TEMPLATE_SYNTAX
#   define CXXTEST_TEMPLATE_INSTANTIATION
#else
#   define CXXTEST_TEMPLATE_INSTANTIATION template<>
#endif

namespace CxxTest 
{
    //
    // This is how we use the value traits
    //
#   define TS_AS_STRING(x) CxxTest::traits(x).asString()

    //
    // The default ValueTraits class just returns "(unprintable value)"
    //
    template <class T>
    class ValueTraits
    {
    public:
	ValueTraits( T ) {}
	const char *asString( void ) const { return "(unprintable value)";}
    };    

    //
    // traits( T t )
    // Creates an object of type ValueTraits<T>
    //
    template <class T>
    inline ValueTraits<T> traits( T t )
    {
	return ValueTraits<T>( t );
    }

    //
    // You can duplicate the implementation of an existing ValueTraits
    //
#   define CXXTEST_COPY_TRAITS(NewClass, OldClass) \
    CXXTEST_TEMPLATE_INSTANTIATION \
    class ValueTraits< NewClass > \
    { \
        ValueTraits< OldClass > _old; \
    public: \
	ValueTraits( NewClass n ) : _old( (OldClass)n ) {} \
	const char *asString( void ) const { return _old.asString(); } \
    };
    
    //
    // Char representation of a digit
    //
    inline char digitToChar( unsigned digit )
    {
	return (digit < 10) ? ('0' + digit) :
	    (digit < 36) ? ('A' + digit - 10) : '?';
    }

    //
    // Avoid compiler warnings about unsigned types always >= 0
    //
    template<class N> inline bool negative( N n ) { return n < 0; }
    template<class N> inline N abs( N n ) { return negative(n) ? -n : n; }

#   define CXXTEST_NON_NEGATIVE(Type) \
    CXXTEST_TEMPLATE_INSTANTIATION \
    inline bool negative<Type>( Type ) { return false; } \
    CXXTEST_TEMPLATE_INSTANTIATION \
    inline Type abs<Type>( Type value ) { return value; }

    CXXTEST_NON_NEGATIVE( bool );
    CXXTEST_NON_NEGATIVE( unsigned char );
    CXXTEST_NON_NEGATIVE( unsigned short int );
    CXXTEST_NON_NEGATIVE( unsigned int );
    CXXTEST_NON_NEGATIVE( unsigned long int );

    //
    // Represent (integral) number as a string
    // Returns one past the end of the string
    // Remember -- we can't use the standard library!
    //
    template<class N>
    inline char *numberToString( N n, char *s,
                                 N base = 10,
				 unsigned skipDigits = 0,
                                 unsigned maxDigits = (unsigned)-1 )
    {
	if ( negative(n) ) {
	    *s++ = '-';
	    n = abs(n);
	}
    
	N digit = 1;
        while ( digit <= (n / base) )
            digit *= base;
	N digitValue;
	for ( ; digit >= 1 && skipDigits; n -= digit * digitValue, digit /= base, -- skipDigits )
	    digitValue = (unsigned)(n / digit);
	for ( ; digit >= 1 && maxDigits; n -= digit * digitValue, digit /= base, -- maxDigits )
	    *s++ = digitToChar( (unsigned)(digitValue = (unsigned)(n / digit)) );

	*s = '\0';
	return s;
    }

    //
    // Copy a string
    // Returns one past the end of the destination string
    //
    inline char *copyString( char *dst, const char *src )
    {
        while ( (*dst = *src) != '\0' ) {
            ++ dst;
            ++ src;
        }
        return dst;
    }

    //
    // Represent a character value as a string
    // Returns one past the end of the string
    // This will be the actual char if printable or '\xXXXX' otherwise
    //
    inline char *charToString( unsigned long c, char *s )
    {
        switch( c ) {
        case '\\': return copyString( s, "\\\\" );
        case '\"': return copyString( s, "\\\"" );
        case '\'': return copyString( s, "\\\'" );
        case '\0': return copyString( s, "\\0" );
        case '\a': return copyString( s, "\\a" );
        case '\b': return copyString( s, "\\b" );
        case '\n': return copyString( s, "\\n" );
        case '\r': return copyString( s, "\\r" );
        case '\t': return copyString( s, "\\t" );
        }
        if ( c >= 32 && c <= 127 ) {
            s[0] = (char)c;
            s[1] = '\0';
            return s + 1;
        }
        else {
            s[0] = '\\';
            s[1] = 'x';
            if ( c < 10 ) {
                s[2] = '0';
                ++ s;
            }
            return numberToString( c, s + 2, 16UL );
        }
    }

    //
    // Prevent negative numbers
    //
    inline char *charToString( char c, char *s )
    {
        return charToString( (unsigned long)(unsigned char)c, s );
    }

    //
    // All the specific ValueTraits follow.
    // You can #define CXXTEST_USER_VALUE_TRAITS if you don't want them
    //
    
#ifndef CXXTEST_USER_VALUE_TRAITS
    //
    // ValueTraits<const char *>
    // This is used for printing strings, as in TS_FAIL( "Message" )
    //
    CXXTEST_TEMPLATE_INSTANTIATION
    class ValueTraits<const char *>
    {
        const char *_asString;
    public:
	ValueTraits( const char *value ) { _asString = value; }
	const char *asString( void ) const { return _asString; }
    };
    
    //
    // ValueTraits<char *>
    // Some compilers need this as well as const char *
    //
    CXXTEST_COPY_TRAITS( char *, const char * );

    //
    // ValueTraits<bool>
    //    
    CXXTEST_TEMPLATE_INSTANTIATION
    class ValueTraits<bool>
    {
        const char *_asString;
    public:
	ValueTraits( bool value ) { _asString = value ? "true" : "false"; }
	const char *asString( void ) const { return _asString; }
    };

    //
    // ValueTraits<signed long>
    //
    CXXTEST_TEMPLATE_INSTANTIATION
    class ValueTraits<signed long>
    {
	typedef long T;
	char _asString[2 + 3 * sizeof(T)];
    public:
	ValueTraits( T t ) { numberToString<T>( t, _asString ); }
	const char *asString( void ) const { return _asString; }
    };

    //
    // ValueTraits<unsigned long>
    //
    CXXTEST_TEMPLATE_INSTANTIATION
    class ValueTraits<unsigned long>
    {
	typedef unsigned long T;
	char _asString[1 + 3 * sizeof(T)];
    public:
	ValueTraits( T t ) { numberToString<T>( t, _asString ); }
	const char *asString( void ) const { return _asString; }
    };

    //
    // All decimals are the same as the long version
    //
    
    CXXTEST_COPY_TRAITS( signed int, signed long int );
    CXXTEST_COPY_TRAITS( unsigned int, unsigned long int );
    
    CXXTEST_COPY_TRAITS( signed short int, signed long int );
    CXXTEST_COPY_TRAITS( unsigned short int, unsigned long int );

    CXXTEST_COPY_TRAITS( unsigned char, unsigned long int );

    //
    // ValueTraits<char>
    // Returns 'x' for printable chars, '\x??' for others
    //
    CXXTEST_TEMPLATE_INSTANTIATION
    class ValueTraits<char>
    {
	char _asString[sizeof("'\\xXX'")];
    public:
	ValueTraits( char c )
        {
            _asString[0] = '\'';
            char *suffix = charToString( c, _asString + 1 );
            suffix[0] = '\'';
            suffix[1] = '\0';
        }
	const char *asString( void ) const { return _asString; }
    };

    //
    // ValueTraits<signed char>
    // Same as char, some compilers need it
    //
    CXXTEST_COPY_TRAITS( signed char, char );

    //
    // ValueTraits<double>
    //
    CXXTEST_TEMPLATE_INSTANTIATION
    class ValueTraits<double>
    {
	enum { maxDigitsOnLeft = 24, digitsOnRight = 4 };
	char _asString[1 + maxDigitsOnLeft + 1 + digitsOnRight + 1];

	static unsigned requiredDigitsOnLeft( double t )
	{
	    unsigned digits = 1;
	    for ( t = (t < 0.0) ? -t : t; t > 1.0; t /= 10.0 )
		++ digits;
	    return digits;
	}

	char *doNegative( double &t )
	{
	    if ( t >= 0 )
		return _asString;
	    _asString[0] = '-';
	    t = -t;
	    return _asString + 1;
	}

	void hugeNumber( double t )
	{
	    char *s = doNegative( t );
	    s = numberToString<double>( t, s, 10, 0, 1 );
	    *s++ = '.';
	    s = numberToString<double>( t, s, 10, 1, digitsOnRight );
	    *s++ = 'E';
	    s = numberToString<double>( requiredDigitsOnLeft( t ) - 1, s );
	}
	
	void normalNumber( double t )
	{
	    char *s = doNegative( t );
	    s = numberToString( t, s );
	    *s++ = '.';
	    for ( unsigned i = 0; i < digitsOnRight; ++ i )
		s = numberToString( (unsigned)(t *= 10.0) % 10, s );
	}
	
    public:
	ValueTraits( double t )
	{
	    (requiredDigitsOnLeft( t ) > maxDigitsOnLeft) ?
		hugeNumber( t ) : normalNumber( t );
	}

	const char *asString( void ) const { return _asString; }
    };

    //
    // ValueTraits<float>
    //
    CXXTEST_COPY_TRAITS( float, double );
#endif // !CXXTEST_USER_VALUE_TRAITS
};

#ifdef CXXTEST_HAVE_STD
#   include <cxxtest/StdValueTraits.h>
#endif // CXXTEST_HAVE_STD


#endif // __CXXTEST__VALUETRAITS_H
