#ifndef __CXXTEST__WIN32GUI_H
#define __CXXTEST__WIN32GUI_H

//
// The Win32Gui displays a simple progress bar using the Win32 API
// It accepts the command line option -minimized
//

#include <cxxtest/Gui.h>

#include <windows.h>
#include <commctrl.h>

namespace CxxTest
{
    class Win32Gui : public GuiListener
    {
    public:
	void enterGui( int &argc, char **argv )
	{
            parseCommandLine( argc, argv );
	}
	
        void enterWorld( unsigned numTotalTests )
        {
            _numTotalTests = numTotalTests;
            _testsDone = 0;
            startGuiThread();
        }

	void enterSuite( const SuiteDescription &d )
	{
	    showSuiteName( d.suiteName() );
	}

        void enterTest( const char *suiteName, const char *testName )
        {
            ++ _testsDone;
            setWindowName( suiteName, testName );
            showTestName( testName );
	    showTestsDone();
            progressBarMessage( PBM_STEPIT );
        }

        void yellowBar()
        {
	    setColor( 255, 255, 0 );
	    setIcon( IDI_WARNING );
        }
        
        void redBar()
        {
            if ( _startMinimized )
                showMainWindow( SW_SHOWNORMAL );
	    setColor( 255, 0, 0 );
	    setIcon( IDI_ERROR );
        }

        void leaveGui()
        {
            DestroyWindow( _mainWindow );
        }

    private:
        const char *_programName;
        bool _startMinimized;
        WNDCLASSEX _windowClass;
        HWND _mainWindow, _progressBar, _statusBar;
        HANDLE _canStartTests;
        unsigned _numTotalTests, _testsDone;
        enum { STATUS_SUITE_NAME, STATUS_TEST_NAME, STATUS_NUM_DONE, STATUS_TOTAL_PARTS };
        int _statusParts[STATUS_TOTAL_PARTS];
        char _statusNumDone[sizeof("1000000000 of 1000000000 (100%)")];

        void parseCommandLine( int argc, char **argv )
        {
            _startMinimized = false;
	    _programName = argv[0];
            
            for ( int i = 1; i < argc; ++ i )
                if ( !lstrcmp( argv[i], "-minimized" ) )
                    _startMinimized = true;
        }
        
        void startGuiThread()
        {
            _canStartTests = CreateEvent( NULL, TRUE, FALSE, NULL );
            CreateThread( NULL, 0, &(Win32Gui::guiThread), (LPVOID)this, 0, NULL );
            WaitForSingleObject( _canStartTests, INFINITE );
        }

        static DWORD WINAPI guiThread( LPVOID parameter )
        {
            ((Win32Gui *)parameter)->gui();
            return 0;
        }

        void gui()
        {
            registerWindowClass();
            createMainWindow();
            initCommonControls();
            createProgressBar();
            createStatusBar();
            centerMainWindow();
            showMainWindow();
            startTests();

            messageLoop();
        }

        void registerWindowClass()
        {
            _windowClass.cbSize = sizeof(_windowClass);
            _windowClass.style = CS_HREDRAW | CS_VREDRAW;
            _windowClass.lpfnWndProc = &(Win32Gui::windowProcedure);
            _windowClass.cbClsExtra = 0;
            _windowClass.cbWndExtra = sizeof(LONG);
            _windowClass.hInstance = (HINSTANCE)NULL;
            _windowClass.hIcon = (HICON)NULL;
            _windowClass.hCursor = (HCURSOR)NULL;
            _windowClass.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
            _windowClass.lpszMenuName = NULL;
            _windowClass.lpszClassName = "CxxTest Window Class";
            _windowClass.hIconSm = (HICON)NULL;

            RegisterClassEx( &_windowClass );
        }

        void createMainWindow()
        {
            _mainWindow = createWindow( _windowClass.lpszClassName, WS_OVERLAPPEDWINDOW );
            SetWindowLong( _mainWindow, GWL_USERDATA, (LONG)this );
        }

        void initCommonControls()
        {
#ifdef ICC_BAR_CLASSES
            INITCOMMONCONTROLSEX init;
            init.dwSize = sizeof(init);
            init.dwICC = ICC_BAR_CLASSES | ICC_PROGRESS_CLASS;
            InitCommonControlsEx( &init );
#else // Old headers, use InitCommonControls()
            InitCommonControls();
#endif // InitCommonControls()/InitCommonControlsEx()
        }

        void createProgressBar()
        {
            _progressBar = createWindow( PROGRESS_CLASS, WS_CHILD | WS_VISIBLE, _mainWindow );

            resizeBars();
#ifdef PBM_SETRANGE32
            progressBarMessage( PBM_SETRANGE32, 0, _numTotalTests );
#else // No PBM_SETRANGE32, use PBM_SETRANGE
	    progressBarMessage( PBM_SETRANGE, 0, MAKELPARAM( 0, _numTotalTests ) );
#endif // PBM_SETRANGE32
            progressBarMessage( PBM_SETPOS, 0 );
            progressBarMessage( PBM_SETSTEP, 1 );
            greenBar();
            UpdateWindow( _progressBar );
        }

        void createStatusBar()
        {
            _statusBar = createWindow( STATUSCLASSNAME, WS_CHILD | WS_VISIBLE, _mainWindow );
        }

        HWND createWindow( LPCTSTR className, DWORD style, HWND parent = (HWND)NULL )
        {
            return CreateWindow( className, NULL, style, 0, 0, 0, 0, parent,
                                 (HMENU)NULL, (HINSTANCE)NULL, (LPVOID)NULL );
        }

        void progressBarMessage( UINT message, WPARAM wParam = 0, LPARAM lParam = 0 )
        {
            SendMessage( _progressBar, message, wParam, lParam );
        }

        void centerMainWindow()
        {
            RECT screen;
            getScreenArea( screen );

            LONG screenWidth = screen.right - screen.left;
            LONG screenHeight = screen.bottom - screen.top;

            LONG xCenter = (screen.right + screen.left) / 2;
            LONG yCenter = (screen.bottom + screen.top) / 2;

            LONG windowWidth = (screenWidth * 4) / 5;
            LONG windowHeight = screenHeight / 10;
            LONG minimumHeight = 2 * (GetSystemMetrics( SM_CYCAPTION ) + GetSystemMetrics( SM_CYFRAME ));
            if ( windowHeight < minimumHeight )
                windowHeight = minimumHeight;

            SetWindowPos( _mainWindow, HWND_TOP,
                          xCenter - (windowWidth / 2), yCenter - (windowHeight / 2),
                          windowWidth, windowHeight, 0 );
        }

        void getScreenArea( RECT &area )
        {
            if ( !getScreenAreaWithoutTaskbar( area ) )
                getWholeScreenArea( area );
        }

        bool getScreenAreaWithoutTaskbar( RECT &area )
        {
            return (SystemParametersInfo( SPI_GETWORKAREA, sizeof(RECT), &area, 0 ) != 0);
        }

        void getWholeScreenArea( RECT &area )
        {
            area.left = area.top = 0;
            area.right = GetSystemMetrics( SM_CXSCREEN );
            area.bottom = GetSystemMetrics( SM_CYSCREEN );
        }

        void showMainWindow()
        {
            showMainWindow( _startMinimized ? SW_MINIMIZE : SW_SHOWNORMAL );
            UpdateWindow( _mainWindow );
        }

        void showMainWindow( int mode )
        {
            ShowWindow( _mainWindow, mode );
        }

        void startTests()
        {
            SetEvent( _canStartTests );
        }

        void messageLoop()
        {
            MSG message;
            while ( BOOL haveMessage = GetMessage( &message, NULL, 0, 0 ) )
                if ( haveMessage != -1 )
                    DispatchMessage( &message );
        }

        static LRESULT CALLBACK windowProcedure( HWND window, UINT message, WPARAM wParam, LPARAM lParam )
        {
            if ( message == WM_SIZE )
                ((Win32Gui *)GetWindowLong( window, GWL_USERDATA ))->resizeBars();
            return DefWindowProc( window, message, wParam, lParam );
        }

        void resizeBars()
        {
            RECT r;
            GetClientRect( _mainWindow, &r );
            LONG width = r.right - r.left;
            LONG height = r.bottom - r.top;

            GetClientRect( _statusBar, &r );
            LONG statusHeight = r.bottom - r.top;
            LONG progressHeight = height - statusHeight;

            resizeBars( width, progressHeight, statusHeight );
            setStatusParts( width );
        }

        void resizeBars( LONG width, LONG progressHeight, LONG statusHeight )
        {
            SetWindowPos( _progressBar, HWND_TOP, 0, 0, width, progressHeight, 0 );
            SetWindowPos( _statusBar, HWND_TOP, 0, progressHeight, width, statusHeight, 0 );
        }

        void setStatusParts( LONG width )
        {
            _statusParts[STATUS_SUITE_NAME] = width / 3;
            _statusParts[STATUS_TEST_NAME] = (width * 5) / 6;
            _statusParts[STATUS_NUM_DONE] = -1;
            statusBarMessage( SB_SETPARTS, STATUS_TOTAL_PARTS, _statusParts );
        }

        void statusBarMessage( UINT message, WPARAM wParam = 0, const void *lParam = 0 )
        {
            SendMessage( _statusBar, message, wParam, (LPARAM)lParam );
        }

        void greenBar()
        {
            setColor( 0, 255, 0 );
            setIcon( IDI_INFORMATION );
        }

        void setColor( BYTE red, BYTE green, BYTE blue )
        {
#ifdef PBM_SETBARCOLOR
            progressBarMessage( PBM_SETBARCOLOR, 0, RGB( red, green, blue ) );
#endif // PBM_SETBARCOLOR
        }

        void setIcon( LPCTSTR icon )
        {
            SendMessage( _mainWindow, WM_SETICON, ICON_BIG, (LPARAM)loadStandardIcon( icon ) );
        }

        HICON loadStandardIcon( LPCTSTR icon )
        {
            return LoadIcon( (HINSTANCE)NULL, icon );
        }

        void setWindowName( const char *suiteName, const char *testName )
        {
            unsigned length = lstrlen( _programName ) + lstrlen( suiteName ) + lstrlen( testName ) + sizeof( " - ::()" );
            char *name = allocate( length );
            wsprintf( name, "%s - %s::%s()", _programName, suiteName, testName );
            SetWindowText( _mainWindow, name );
            free( name );
        }

        void showSuiteName( const char *suiteName )
        {
            setStatusPart( STATUS_SUITE_NAME, suiteName );
	}

	void showTestName( const char *testName )
	{
            setStatusPart( STATUS_TEST_NAME, testName );
	}

	void showTestsDone()
	{
            wsprintf( _statusNumDone, "%u of %u (%u%%)",
                      _testsDone, _numTotalTests,
                      (_testsDone * 100) / _numTotalTests );
            setStatusPart( STATUS_NUM_DONE, _statusNumDone );
        }

        void setStatusPart( UINT part, const char *text )
        {
            statusBarMessage( SB_SETTEXT, part, text );
        }

        char *allocate( unsigned length )
        {
            return (char *)HeapAlloc( GetProcessHeap(), 0, length );
        }

        void free( char *data )
        {
            HeapFree( GetProcessHeap(), 0, data );
        }
    };
};

#endif // __CXXTEST__WIN32GUI_H
