#!/usr/bin/perl -w

#
# This script does a basic "black-box" test of CxxTest
# Run it with the --no-python option test only cxxtestgen.pl
#

use strict;
use File::Compare;
use Getopt::Long;
use English;

# Automatic flush of STDOUT
$OUTPUT_AUTOFLUSH = 1;

#
# Some command line options
#
my $noPython = 0;
my $noPerl = 0;
my $compiler;
my $compilee;
my $exe_option = '-o ';
my $include_option = '-I';
my ($quiet, $dots);
GetOptions( 'no-python' => \$noPython,
	    'no-perl' => \$noPerl,
	    'compiler=s' => \$compiler,
	    'exe-option=s' => \$exe_option,
	    'include-option=s' => \$include_option,
	    'quiet' => \$quiet,
	    'dots' => \$dots,
	  );

#
# --no-python --no-perl is meaningless
#
if ( $noPerl && $noPython ) {
  die "You can specify at most one of --no-perl and --no-python\n";
}

$compilee = $noPerl ? "py.cpp" : "pl.cpp";

#
# Simplistic compiler selection
#
if ( !$compiler ) {
  $compiler = ("$^O" eq 'MSWin32') ? 'cl -nologo -GX' : 'c++';
}
my $target = ("$^O" eq 'MSWin32') ? 'px.exe' : './px';
my $compile = "${compiler} ${exe_option}${target} ${include_option}. ${include_option}.. ${compilee} > build.log 2>&1";
my $compileWithMain = "${compiler} ${exe_option}${target} ${include_option}.. main.cpp ${compilee} > build.log 2>&1";

#
# Compare two files line by line
# This sub leaves the two files open
#
sub _filesEqual($$) {
  open FILE1, "<$_[0]" or die "Missing file $_[0]\n";
  open FILE2, "<$_[1]" or die "Missing file $_[1]\n";
  while (<FILE1>) {
    if ( $_ ne <FILE2> ) {
      return 0;
    }
  }
  return !<FILE2>;
}

#
# Compare two files line by line
#
sub filesEqual($$) {
  my $result = _filesEqual($_[0], $_[1]);
  close FILE1;
  close FILE2;
  return $result;
}

#
# Print something (unless we were told to keep quiet)
#
sub say(@) {
  if ( $quiet ) {
  }
  elsif ( $dots ) {
    print ".";
  }
  else {
    print @_;
  }
}

#
# Append a line to a file that reports the error level
#
sub appendErrorLevel() {
  my $errorLevel = $? >> 8;
  open FILE, ">>px.out";
  print FILE "Error level = $errorLevel\n";
  close FILE;
}

#
# Run one test case:
#  - Generate source file(s)
#  - Optionally compile
#  - Compare outputs to reference file
#
sub testCase($) {
  my $test = $_[0];
  my ($description, $output, $args) = ($test->{'description'}, $test->{'output'}, $test->{'args'});
  clean();
  say("$description: ");

  if ( !$noPerl ) {
    say "perl...";
    system( "perl ../cxxtestgen.pl -o pl.cpp $args > pl.out 2>&1" );
  }

  if ( !$noPython ) {
    say "python...";
    system( "python ../cxxtestgen.py -o py.cpp $args > py.out 2>&1" );
  }

  if ( $output && "$output" ne "" ) {
    compileAndCheckOutput( $test );
  } else {
    checkNoSources();
  }

  say "OK\n";
  clean();
}

#
# Compile generated source file, run it and verify the output
#
sub compileAndCheckOutput($) {
  my ($test) = @_;

  if ( !$noPython && !$noPerl ) {
    say "compare...";
    if ( !filesEqual( "pl.cpp", "py.cpp" ) ) {
      die "Different source files generated (pl.cpp <> py.cpp)\n";
    }
  }

  say "compile...";
  system( $test->{'main'} ? "$compileWithMain" : "$compile" );

  say "run...";
  if (my $run = $test->{'run'}) {
    system( "$run" );
  } else {
    system( "$target > px.out 2>&1" );
  }
  appendErrorLevel();

  say "check...";
  my $output = $test->{'output'};
  if ( !filesEqual( $output, "px.out" ) ) {
    die "Generated source gives incorrect output ($output <> px.out)\n";
  }
}

#
# Verify that no source files have been generated
#
sub checkNoSources() {
  if ( !$noPerl && -e "pl.cpp" ) {
    die "Output file pl.cpp should not have been generated\n";
  }
  if( !$noPython && -e "py.cpp" ) {
    die "Output file py.cpp should not have been generated\n";
  }
}

#
# Check for basic_string<wchar_t> support
#
sub haveWideChar() {
  system( "${compiler} ${exe_option}${target} ${include_option}.. wchar.cpp > build.log 2>&1" );
  return ( -x $target );
}

#
# Clean up the mess
#
sub clean() {
  unlink "pl.cpp", "py.cpp", "pl.out", "py.out", "$target", "px.out", "build.log", <*~>;
  checkNoSources();
}

#
# Make sure we scan the sample headers in the correct order
#
my $samples = join( " ", sort( <../sample/*.h> ) );

#
# Run all the test cases
#
testCase { 'description' => "No Errors       ", 'args' => "--error-printer GoodSuite.h", 'output' => "good.out" };
testCase { 'description' => "One input       ", 'args' => "../sample/SimpleTest.h", 'main' => 1, 'output' => "one.out" };
testCase { 'description' => "Wildcard input  ", 'args' => '../sample/*.h', 'main' => 1, 'output' => "wildcard.out" };
testCase { 'description' => "Error Printer   ", 'args' => "--error-printer $samples", 'output' => "error.out" };
testCase { 'description' => "Stdio Printer   ", 'args' => "--runner=StdioPrinter $samples", 'output' => "error.out" };
testCase { 'description' => "Paren Printer   ", 'args' => "--runner=ParenPrinter $samples", 'output' => "paren.out" };
testCase { 'description' => "Runner option   ", 'args' => "--runner=YesNoRunner $samples", 'output' => "runner.out" };
testCase { 'description' => "ValueTraits     ", 'args' => "--error-printer TraitsTest.h", 'output' => "traits.out" };
if ( haveWideChar() ) {
  testCase { 'description' => "Wide Char       ", 'args' => "--error-printer WideCharTest.h", 'output' => "wchar.out" };
}
testCase { 'description' => "Template        ", 'args' => "--template=../sample/file_printer.tpl $samples",
	     'run' => "$target px.out", 'output' => "error.out" };
testCase { 'description' => "User Traits     ", 'args' => "--template=UserTraits.tpl UserTraits.h", 'output' => 'user.out' };
testCase { 'description' => "Abort on fail   ", 'args' => "--error-printer --abort-on-fail $samples", 'output' => "aborter.out" };
testCase { 'description' => "Aborter w/o EH  ", 'args' => "--error-printer --have-eh --abort-on-fail AborterNoThrow.h", 'output' => "aborter_no_throw.out" };
testCase { 'description' => "Throw w/o Std   ", 'args' => "--template=ThrowNoStd.tpl ThrowNoStd.h", 'output' => 'throw.out' };
testCase { 'description' => "GUI             ", 'args' => "--gui=DummyGui GreenYellowRed.h", 'output' => "gui.out" };
testCase { 'description' => "GUI + Runner    ", 'args' => "--gui=DummyGui --runner=ParenPrinter GreenYellowRed.h", 'output' => "gui_paren.out" };
testCase { 'description' => "Missing destroy ", 'args' => "BadTest.h" };
testCase { 'description' => "No tests        ", 'args' => "EmptySuite.h" };
testCase { 'description' => "Missing Input   ", 'args' => "NoSuchFile.h" };
testCase { 'description' => "Missing Template", 'args' => "--template=NoSuchFile.h $samples" };

#
# Local Variables:
# compile-command: "perl test.pl"
# End:
#
